# Package init file

'''LXCA Python Client API'''

#setting CMD_PATH environment variable to be used in submodules for loading data files
import os
pylxca_api_path = os.path.dirname(__file__)
os.environ['PYLXCA_API_PATH'] = pylxca_api_path

# All submodules of this package are imported; so clients need to import just this package.
from .lxca_api import *


'''
@since: 15 Sep 2015
@author: Girish Kumar <gkumar1@lenovo.com>, Prashant Bhosale <pbhosale@lenovo.com>
@license: Lenovo License
@copyright: Copyright 2016, Lenovo
@organization: Lenovo 
@summary: This module provides interactive shell class which provides base for interactive shell 
feature and accepts raw input from command line 
'''

import re,sys,os,traceback
import itertools
import logging
import shlex
from platform import python_version

from pylxca.pylxca_cmd import lxca_cmd
from pylxca.pylxca_cmd.lxca_icommands import InteractiveCommand
from pylxca.pylxca_cmd.lxca_view import lxca_ostream
from pylxca.pylxca_cmd import lxca_icommands
#from rlcompleter import readline

logger = logging.getLogger(__name__)

class InteractiveShell(object):

    class help(InteractiveCommand):
        """
        Prints a list of available commands and their descriptions, or the help
        text of a specific command. Requires a list of the available commands in
        order to display text for them. 
        
        @ivar commands: A dictionary of available commands, bound to L{InteractiveShell.commands}
        @type commands: C{dict}
        """
        def __init__(self, commands, shell=None):
            """
            Constructor function for L{_HelpCommand}.
            
            @param commands: A dictionary of available commands, usually L{InteractiveShell.commands}
            @type commands: C{dict}
            """
            self.commands = commands
            self.shell = shell

        def handle_command(self, opts, args):
            """
            Prints a list of available commands and their descriptions if no
            argument is provided. Otherwise, prints the help text of the named
            argument that represents a command. Does not throw an error if the
            named argument doesn't exist in commands, simply prints a warning.
            
            @param opts: Will be an empty dictionary
            @type opts: C{dict}
            @param args: The raw string passed to the command, either a command or nothing
            @type args: C{list}
            
            @return: Returns nothing, sends messages to stdout
            @rtype: None
            """
            
            if len(args) == 0:
                self.do_command_summary( )
                return

            if args[0] not in self.commands:
                self.sprint('No help available for unknown command "%s"' % args[0])
                return

            self.commands[args[0]].get_argparse_options().print_help()

        def do_command_summary(self):
            """
            If no command is given to display help text for specifically, then
            this helper method is called to print out a list of the available
            commands and their descriptions. Iterates over the list of commands,
            and gets their summary from L{lxca_icommands.get_short_desc}
            
            @return: Returns nothing, sends messages to stdout
            @rtype: C{None}
            """
            self.sprint('The following commands are available:\n')

            cmdwidth = 0
            for name in list(self.commands.keys( )):
                if len(name) > cmdwidth:
                    cmdwidth = len(name)

            cmdwidth += 2
            for name in sorted(self.commands.keys( )):
                command = self.commands[name]

                if name == 'help':
                    continue
                try:
                    self.sprint('  %s   %s' % (name.ljust(cmdwidth),
                                     command.get_short_desc( )))
                except:
                    #For some commands there is no short desc so it will be skipped
                    continue

        def get_short_desc(self):
            return ''
    
        def get_help_message(self):
            return ''

    class exit(InteractiveCommand):
        """
        Exits the interactive shell. 
        """

        def handle_command(self, opts, args):
            sys.exit(0)

        def get_short_desc(self):
            return 'Exit the program.'
    
        def get_help_message(self):
            return 'Type exit and the program will exit.  There are no options to this command'


    def __init__(self, 
        banner="Welcome to Interactive Shell",
        prompt=" >>> "):
        self.banner = banner
        self.prompt = prompt
        self.commands = { }
        self.ostream = lxca_ostream()
        self.add_command(self.help(self.commands,self))
        self.add_command(self.exit())
        
        self.add_command(lxca_cmd.connect(self))
        self.add_command(lxca_cmd.disconnect(self))
        self.add_command(lxca_cmd.log(self))
        self.add_command(lxca_cmd.chassis(self))
        self.add_command(lxca_cmd.nodes(self))
        self.add_command(lxca_cmd.switches(self))
        self.add_command(lxca_cmd.fans(self))
        self.add_command(lxca_cmd.powersupplies(self))
        self.add_command(lxca_cmd.fanmuxes(self))
        self.add_command(lxca_cmd.cmms(self))
        self.add_command(lxca_cmd.scalablesystem(self))
        self.add_command(lxca_cmd.ostream(self))
        #self.add_command(lxca_cmd.jobs(self))
        self.add_command(lxca_cmd.discover(self))
        self.add_command(lxca_cmd.manage(self))
        self.add_command(lxca_cmd.unmanage(self))
        self.add_command(lxca_cmd.configpatterns(self))
        self.add_command(lxca_cmd.configprofiles(self))
        self.add_command(lxca_cmd.configtargets(self))
        self.add_command(lxca_cmd.users(self))
        self.add_command(lxca_cmd.lxcalog(self))
        self.add_command(lxca_cmd.ffdc(self))
        self.add_command(lxca_cmd.updatepolicy(self))
        self.add_command(lxca_cmd.updatecomp(self))
        self.add_command(lxca_cmd.updaterepo(self))
        self.add_command(lxca_cmd.tasks(self))
        self.add_command(lxca_cmd.manifests(self))
        self.add_command(lxca_cmd.osimages(self))
        self.add_command(lxca_cmd.resourcegroups(self))
        self.add_command(lxca_cmd.managementserver(self))
        self.add_command(lxca_cmd.rules(self))
        self.add_command(lxca_cmd.compositeResults(self))
        self.add_command(lxca_cmd.storedcredentials(self))
        self.add_command(lxca_cmd.license(self))

    def set_ostream_to_null(self):
        self.ostream = open(os.devnull, 'w')
        
    def sprint(self,str):
        if self.ostream:
            self.ostream.write(str)
    
    def print_Hello(self):
        self.sprint("Hello")
        
    def add_command(self, command):
        if command.get_name( ) in self.commands:
            raise Exception('command %s already registered' % command.get_name( ))

        self.commands[command.get_name()] = command

    def run(self):
        self.sprint('')
        self.sprint('-'*50)
        self.sprint(self.banner)
        self.sprint('Type "help" at any time for a list of commands.')
        self.sprint('Use "lxca_shell --api" to enable Interactive Python LXCA Shell ')
        self.sprint('-'*50)
        self.sprint('')
            
        while True:
            try:
#                readline.set_completer_delims(' \t\n;')
#                readline.parse_and_bind("tab: complete")
#                readline.set_completer(self.auto_complete)
                if python_version()[0] == '2':
                    command_line = raw_input(self.prompt)
                else:
                    command_line = input(self.prompt)
            except (KeyboardInterrupt, EOFError):
                break

            self.handle_input(command_line)
            continue
            
    # TODO: Need to rename this function to handle_cli_params
    def handle_input(self, command_line):

        command_line = command_line.strip()

        # Get command
        command_args = ""
        if " " in command_line:
            command_line_index = command_line.index(" ")
            command_name = command_line[:command_line_index]
            command_args = command_line[command_line_index+1:]
        else:
            command_name = command_line

        # Show message when no command
        if not command_name:
            return

        if not command_name in self.commands:
            self.sprint('Unknown command: "%s". Type "help" for a list of commands.' % command_name)
            return 

        command = self.commands[command_name]

        opts = {}
        args = []

        # Split the input to allow for quotes option values
        #re_args = re.findall('\-\-\S+\=\"[^\"]*\"|\S+', command_args)
        re_args = shlex.split(command_args)
        #re_args = re.findall('\-\-\S+|\-\S+|\S+[\s*\w+]*', command_args)
        # Parse args if present
        for i in range(0, len(re_args)):
            args.append( re_args[i] )
        try:
            return command.handle_command(opts=opts, args=args)
        except Exception as err:
            logger.error("Exception occurred while processing command %s " %str(err))
        return
    
    # TODO: We should remove this, All API level code should give call to handle_input_dict rather
    def handle_input_args(self, command_name,*args, **kwargs):
        # Show message when no command
        if not command_name:
            return

        if not command_name in self.commands:
            self.sprint('Unknown command: "%s". Type "help" for a list of commands.' % command_name)
            return

        command = self.commands[command_name]

        opts = {}
        args_dict = {}
       
        for item in kwargs['kwargs']:
            args_dict['--' + item] = kwargs['kwargs'][item]

        args_list = list(args)

        args_list = args_list + list(itertools.chain(*list(args_dict.items())))
        try:
            return command.handle_command(opts=opts, args=args_list)
        except Exception as err:
            self.sprint("Exception occurred while processing command.")
            raise err
            
    # TODO: We should rename this to handle_api_params
    def handle_input_dict(self, command_name,con, param_dict, use_argparse = True):
        # Show message when no command
        if not command_name:
            return

        if not command_name in self.commands:
            self.sprint('Unknown command: "%s". Type "help" for a list of commands.' % command_name)
            return

        command = self.commands[command_name]

        if use_argparse:
            try:
                args_list = []
                args_dict = {}
                if 'subcmd' in param_dict:
                    subcmd = param_dict['subcmd']
                    param_dict.pop('subcmd')
                    args_list.append(subcmd)

                for item in param_dict:
                    args_dict['--' + item] = param_dict[item]

                args_list = args_list + list(itertools.chain(*list(args_dict.items())))
                param_dict = command.parse_args(args_list)
            except SystemExit as err:
                self.sprint("Exception occurred while parsing command.")
                raise Exception("Invalid argument")
            except Exception as err:
                self.sprint("Exception occurred while parsing api input.")
                raise err

        try:
            return command.handle_input(param_dict, con)
        except Exception as err:
            self.sprint("Exception occurred while processing command.")
            raise err


    """
    TODO:
    def auto_complete(self, text, state):
        command_list = self.commands.keys()
        re_space = re.compile('.*\s+$', re.M)
        "Generic readline completion entry point."
        buffer = readline.get_line_buffer()
        line = readline.get_line_buffer().split()
        # show all commands
        if not line:
            return [c + ' ' for c in command_list][state]
        # account for last argument ending in a space
        if re_space.match(buffer):
            line.append('')
        # resolve command to the implementation function
        cmd = line[0].strip()
        if cmd in command_list:
            impl = getattr(self, 'complete_%s' % cmd)
            args = line[1:]
            if args:
                return (impl(args) + [None])[state]
            return [cmd + ' '][state]
        results = [c + ' ' for c in command_list if c.startswith(cmd)] + [None]
        return results[state]
    """


'''
@since: 15 Sep 2015
@author: Girish Kumar <gkumar1@lenovo.com>, Prashant Bhosale <pbhosale@lenovo.com>
@license: Lenovo License
@copyright: Copyright 2016, Lenovo
@organization: Lenovo 
@summary: This module provides interactive command class which provides base 
implementation for all command classes 
'''

import sys,getopt,os,json,logging, traceback
import argparse
from pylxca.pylxca_api import lxca_api
from pylxca.pylxca_api.lxca_rest import HTTPError
from pylxca.pylxca_api.lxca_connection import ConnectionError
from pylxca.pylxca_cmd import lxca_view
import textwrap
import ast
cmd_data_json_file   = "lxca_cmd_data.json"
pylxca_cmd_data   = os.path.join(os.getenv('PYLXCA_CMD_PATH'), cmd_data_json_file)

class InteractiveCommand(object):
    def __init__(self, shell=None ):
        self.shell = shell
        fp = open(pylxca_cmd_data, 'r')
        self.command_data = json.load(fp)

    def get_options(self):
        return {}

    def get_name(self):
        return self.__class__.__name__
    
    def get_help_option(self):
        return self.command_data[self.__class__.__name__].get('add_help', True)
    
    def _process_epilog(self, str_list):
        epilog = "\r\n ".join(str_list)
        epilog = textwrap.dedent(epilog)
        return epilog

    def _validate_combination(self, input_dict, valid_dict):
        '''
        This function validate input combinations
        :param input_dict:  input dict of parameters
        :param valid_list:  valid combination of parameters list
        :return: if True returns True and choosed combination of parameter
                 on failure return false and suggested_combination of parameters
        '''

        try:
            if 'subcmd' in input_dict:
                valid_list = valid_dict[input_dict['subcmd']]
            else:
                valid_list = valid_dict['global']
        except Exception as e:
            raise Exception("Error in getting valid_list")

        # remove None and empty string and len zero lists
        # create copy dict
        copy_input_dict = {}
        for k in input_dict.keys():
            if k not in ['func', 'view','subcmd']:
                if not (input_dict[k] == None or (isinstance(input_dict[k], list) and len(input_dict[k]) == 0)):
                    copy_input_dict[k] = input_dict[k]

        input_key_set = set(copy_input_dict.keys())

        for comb in valid_list:
            comb_set = set(comb)
            if input_key_set == comb_set:
                return True, comb

        # Check for suggestion
        suggested_combination = ""
        for comb in valid_list:
            comb_set = set(comb)
            if len(input_key_set & comb_set) > 0:
                suggested_combination += str(comb)

        return False, suggested_combination

    def get_additional_detail(self):
        epilog = self.command_data[self.__class__.__name__].get('additional_detail', [])
        return self._process_epilog(epilog)

    def get_short_desc(self):
        return self.command_data[self.__class__.__name__]['description']

    def post_parsing_validation(self, opts):
        valid_combination = self.command_data[self.__class__.__name__].get('valid_combination', None)

        if valid_combination:
            valid, combination = self._validate_combination(opts, valid_combination)
            if not valid:
                raise Exception("Invalid Missing Arguments %s" % str(combination))

    def cmd1(self, args):
        print('cmd1', args)

    def _preprocess_argparse_dict(self, cmd_dict):
        try:
            replace_with = {"int": int, "boolean": bool, "json.loads":json.loads, "ast.literal_eval":ast.literal_eval}
            if 'type' in cmd_dict:
                if cmd_dict['type'] in replace_with.keys():
                    cmd_dict['type'] = replace_with[cmd_dict['type']]
        except Exception as err:
            raise("Error while preprocessing  argparse dict for type replacement")
    def get_argparse_options(self):
        parser = argparse.ArgumentParser(prog=self.get_name(), description=self.get_short_desc(),
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         epilog=self.get_additional_detail(),
                                         add_help=self.get_help_option())

        arg_list = self.command_data[self.__class__.__name__].get('cmd_args', None)
        if arg_list:
            for opt in arg_list:
                cmd_args = opt[0]["args"]
                cmd_args_list = cmd_args.split(",")
                cmd_args_tuple = tuple(cmd_args_list)
                cmd_dict = opt[0]["opt_dict"]
                self._preprocess_argparse_dict(cmd_dict)
                parser.add_argument(*cmd_args_tuple, **cmd_dict)

        # Add mutually exclusive args if any specified
        arg_list = self.command_data[self.__class__.__name__].get('mutually_exclusive_args', None)
        if arg_list:
            group = parser.add_mutually_exclusive_group()
            for opt in arg_list:
                cmd_args = opt[0]["args"]
                cmd_args_list = cmd_args.split(",")
                cmd_args_tuple = tuple(cmd_args_list)
                cmd_dict = opt[0]["opt_dict"]
                self._preprocess_argparse_dict(cmd_dict)
                group.add_argument(*cmd_args_tuple, **cmd_dict)

        # Add subcommand if any specified
        subcmd_list = self.command_data[self.__class__.__name__].get('subcmd', None)
        if subcmd_list:
            sub_parser = parser.add_subparsers(dest='subcmd')
            sub_parser.required = True
            for subcmd in subcmd_list:
                sub_cmd_name = subcmd['name']
                subcmd_args_list = subcmd['subcmd_args']
                parser_internal = sub_parser.add_parser(sub_cmd_name,
                                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                                        help=subcmd['help'], epilog=self._process_epilog(subcmd['additional_detail']))
                parser_internal.required = True
                parser_internal.set_defaults(func=self.cmd1)
                for opt in subcmd_args_list:
                    cmd_args = opt[0]["args"]
                    cmd_args_list = cmd_args.split(",")
                    cmd_args_tuple = tuple(cmd_args_list)
                    cmd_dict = opt[0]["opt_dict"]
                    self._preprocess_argparse_dict(cmd_dict)
                    parser_internal.add_argument(*cmd_args_tuple, **cmd_dict)

                # Add mutually exclusive args if any specified
                arg_list = subcmd.get('mutually_exclusive_args', None)
                if arg_list:
                    group = parser_internal.add_mutually_exclusive_group()
                    for opt in arg_list:
                        cmd_args = opt[0]["args"]
                        cmd_args_list = cmd_args.split(",")
                        cmd_args_tuple = tuple(cmd_args_list)
                        cmd_dict = opt[0]["opt_dict"]
                        self._preprocess_argparse_dict(cmd_dict)
                        group.add_argument(*cmd_args_tuple, **cmd_dict)

        return parser

    def invalid_input_err(self):
        self.sprint("Invalid Input ")
        self.sprint("for help type command -h")
        return
    
    def sprint(self,str):
        if self.shell: self.shell.sprint(str)

    def parse_args(self, args):
        try:
            parser = self.get_argparse_options()
            namespace = parser.parse_args(args)
            opt_dict = vars(namespace)
            self.post_parsing_validation(opt_dict)
        except argparse.ArgumentError as e:
            raise(e)
        except SystemExit as e:
            raise(e)
        except AttributeError as e:
            #TODO  move this some where
            extype, ex, tb = sys.exc_info()
            formatted = traceback.format_exception_only(extype, ex)[-1]
            message = "Check getopt short and long options  %s" % (formatted)
            raise RuntimeError(message, tb)
        return opt_dict

    def handle_no_input(self,con_obj):
        #no_opt action can differ command to command so override this function if required
        obj = None
        try:
            api = lxca_api()
            obj = api.api(self.get_name(), None,con_obj)
        except ConnectionError:
            self.sprint("Connection is not Initialized, Try connect")
        except RuntimeError:
            self.sprint("Session Error to LXCA, Try connect")
        except Exception as err:
            self.sprint("Exception occurred: %s" %(err)) 
        return obj
    
    def handle_input(self, dict_handler,con_obj = None):
        obj = None
        api = lxca_api()
        obj = api.api(self.get_name(),dict_handler,con_obj)
        return obj
    
    def show_output(self, py_obj,view_filter = "default"):

        ostream = sys.__stdout__
        if self.shell:
            ostream = self.shell.ostream
        view = lxca_view.lxca_view(ostream)
        view.show_output(py_obj,self.get_name(),view_filter)
        return
    
    def handle_output(self, py_obj):
        return
    
    def handle_command(self,  opts, args):
        
        con_obj = None

        try:
            opts = self.parse_args(args)
        except SystemExit as e:
            # ignore this as we need to continue on shell
            return

        out_obj = None
        opt_dict = None
        view_filter = "default"
        
        try:
            if not opts:
                out_obj = self.handle_no_input(con_obj)
            else:
                #opt_dict = self.parse_args(opts, argv)

                out_obj = self.handle_input(opts,con_obj)
                if opts:
                    view_filter = next((item for item in [opts.get('v') , opts.get('view')] if item is not None),'default')
        
            if out_obj:
                if isinstance(out_obj, dict):
                    self.show_output(out_obj,view_filter)
                else:
                    self.handle_output(out_obj)
        except ConnectionError:
            self.sprint("Connection is not Initialized, Try connect")
        except HTTPError as re:
            self.sprint("Exception %s occurred while executing command."%(re.response.content))
        except ConnectionError as re:
            self.sprint("Exception %s occurred while executing command."%(re.response.content))
        except RuntimeError:
            self.sprint("Session Error to LXCA, Try connect")
        except Exception as err:
            self.sprint("Exception occurred: %s" %(err)) 
            
        return out_obj



'''
@since: 21 Oct 2015
@author: Prashant Bhosale <pbhosale@lenovo.com>
@license: Lenovo License
@copyright: Copyright 2016, Lenovo
@organization: Lenovo
@summary: This module is for view logic of the console commands. It parses the dictionary
data and displays on ostream.
'''

import json, re, os, sys, logging
from pprint import pprint
import xml.etree.cElementTree as ElementTree

import pylxca.pylxca_cmd

filter_file = "lxca_filters.xml"
output_file = "lxca_console.out"
pylxca_filter = os.path.join(os.getenv('PYLXCA_CMD_PATH'), filter_file)
pylxca_outfile = os.path.join(os.getenv('PYLXCA_CMD_PATH'), output_file)

indent = 0
logger = logging.getLogger(__name__)

class Tee(object):
    def __init__(self, *files):
        self.files = files
    def write(self, obj):
        for f in self.files:
            f.write(obj)
            f.flush() # If you want the output to be visible immediately
    def flush(self) :
        for f in self.files:
            f.flush()

class lxca_ostream(object):
    def __init__(self):
        self.stdout = sys.__stdout__
        self.print_lvl = 1

    def get_lvl(self):
        return self.print_lvl

    def set_lvl(self,lvl):
        self.print_lvl = lvl

        try:
            if lvl == 0:
                self.stdout = open(os.devnull, 'w')
            elif lvl == 1:
                self.stdout = sys.__stdout__
            elif lvl == 2:
                self.stdout = open(pylxca_outfile, 'w')
            elif lvl == 3:
                outfile = open(pylxca_outfile, 'w')
                self.stdout = Tee(sys.__stdout__, outfile)
        except Exception as e:
            return False
        return True

    def write(self, string):
        sys.stdout = self.stdout
        print(string)
        sys.stdout = sys.__stdout__

class lxca_view(object):

    def __init__(self,ostream = sys.__stdout__):
        self.ostream = ostream
        self.vf_dict =  { 'chassis':'chassisList',
                          'nodes':'nodesList',
                          'switches':'switchList',
                          'fans':'fanList',
                          'powersupplies':'powerSupplyList',
                          'fanmuxes':'fanMuxList',
                          'cmms':'cmmList',
                          'scalablesystem':'scalablesystem',
                          'discovery':'discovery',
                          'updatepolicy': 'policies'}


    def get_val(self,py_obj, tag ):
        a = []
        try:
            if isinstance(py_obj, (dict)):
                return(py_obj[tag])
            if isinstance(py_obj, (list)):

                for i in range(0, len(py_obj)):
                    a.append(py_obj[i][tag])
        except:
            return None
        return a

    def get_view_filter(self, cmd_name, filter_tag):
        vf_tree = ElementTree.parse(pylxca_filter)
        vf_root = vf_tree.getroot()

        for vf in vf_root.findall(cmd_name):
            if vf.attrib['name']==filter_tag:
                return vf

    def print_recur(self,py_obj,view_filter):
        """Recursively prints the python object content as per view filter"""
        global indent
        if str(view_filter.attrib.get('type')) != "object" :
            self.ostream.write(' '*indent + '%s: %s' % (view_filter.tag.title(), self.get_val(py_obj,view_filter.attrib.get('name', view_filter.text))))
        #else:
        #    self.ostream.write('%s: ' % (view_filter.tag.title()))

        indent += 4
        # View Filter has children so
        py_obj_item = self.get_val(py_obj, view_filter.attrib.get('name', view_filter.text))
        #if py_obj_item is list then iterate through the list and call print recur for each
        if isinstance(py_obj_item, (list)):
            for item in py_obj_item:
                for elem in view_filter.iter():
                    self.print_recur(item,elem)
        else:
            for elem in view_filter.iter():
                self.print_recur(py_obj_item,elem)
        indent -= 4

    def print_cmd_resp_object(self,cmd_resp_item, vf):
        for vf_elem in vf.iter():
            self.print_recur(cmd_resp_item,vf_elem)


    def show_output(self,cmd_reponse, cmd_name,filter_tag):
        vf = self.get_view_filter(cmd_name,filter_tag)
        self.ostream.write("Printing "+ cmd_name + " Output:"+ "\n")

        if len(list(cmd_reponse.keys())) == 0:
            self.ostream.write("No "+ filter_tag + " returned."+ "\n")
        elif len(list(cmd_reponse.keys())) > 1:
            self.print_cmd_resp_object(cmd_reponse, vf)
        else:
            for cmd_resp_item in cmd_reponse[list(cmd_reponse.keys())[0]]:
                self.print_cmd_resp_object(cmd_resp_item, vf)
                self.ostream.write('\n-----------------------------------------------------')


'''
@since: 15 Sep 2015
@author: Prashant Bhosale <pbhosale@lenovo.com>
@license: Lenovo License
@copyright: Copyright 2016, Lenovo
@organization: Lenovo 
@summary: This module provides command class implementation for PyLXCA 
'''

import sys, getopt
import logging, traceback
from getpass import getpass
import json
import argparse

import pylxca.pylxca_api
from pylxca.pylxca_api.lxca_rest import HTTPError
from pylxca.pylxca_api.lxca_connection import ConnectionError
from pylxca.pylxca_cmd.lxca_icommands import InteractiveCommand

logger = logging.getLogger(__name__)

class connect(InteractiveCommand):
    """
    Connects to the LXCA Interface

    USAGE:
        connect -h | --help
        connect -l <URL> -u <USER> [--noverify]
    
    OPTIONS:
        -h        This option displays command usage information
        -l, --url    URL of LXCA
        -u, --user    Username to authenticate
        --noverify    Do not verify the server certificate for https URLs

    """
    def handle_command(self, opts, args):

        '''
        try:
            parser = self.get_argparse_options()
            namespace = parser.parse_args(args)
        except argparse.ArgumentError as e:
            self.invalid_input_err()
            return
        except SystemExit as e:
            # -h and --help land here
            return


        for opt, arg in opts:
            if '-h' in opt:
                self.sprint (self.__doc__)
                return                

        if not opts:
            self.handle_no_input()
            return
        
        opt_dict = self.parse_args(opts, argv)
        '''

        try:
            opt_dict = self.parse_args(args)
            if opt_dict.get('pw', None) == None:
                opt_dict ['pw'] = getpass("Enter Password: ")
        except SystemExit as e:
            # ignore this as we need to continue on shell
            return

        out_obj = None
        
        try:
            out_obj = self.handle_input(opt_dict)
            self.handle_output(out_obj)
        except HTTPError as re:
            self.sprint("Exception %s occurred while executing command."%(re))
        except ConnectionError as re:
            self.sprint("Exception %s occurred while executing command."%(re))
        except RuntimeError:
            self.sprint("Session Error to LXCA, Try connect")
        except Exception as err:
            self.sprint("Exception occurred: %s" %(err)) 
    
        return out_obj
    
    def handle_no_input(self,con_obj = None):
        #no_opt action can differ command to command so override this function if required
        self.invalid_input_err()
        return
    
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Failed to connect given LXCA " )
        else:
            self.sprint("Connection to LXCA successful")
        return
    
###############################################################################
    
class disconnect(InteractiveCommand):
    """
    Diconnects from LXCA Interface 

    USAGE:
        disconnect -h | --help
        disconnect
    """
    def handle_no_input(self, con_obj = None):
        api = pylxca.pylxca_api.lxca_api()
        if api.disconnect() == True:
            self.sprint("Connection with LXCA closed successfully " )
        else:
            self.sprint("Failed to close connection with LXCA " )
        return 

###############################################################################

class log(InteractiveCommand):
    """
    Retrieve and configure logging of LXCA Python tool
    
    USAGE:
        log -h | --help
        log [-l <level>]

    OPTIONS:
        -l, --lvl    Logging level
      
    """
    def handle_no_input(self,con_obj = None):
        api = pylxca.pylxca_api.lxca_api()
        self.sprint("Current Log Level is set to " + str(api.get_log_level()))
        message = """
Possible Log Levels, Please use following values to set desired log level. 

\tDEBUG:        Detailed information, typically of interest only when diagnosing problems.
\tINFO:        Confirmation that things are working as expected.
\tWARNING:    An indication that something unexpected happened, or indicative of some problem in the near future. 
\tERROR:        Due to a more serious problem, the software has not been able to perform some function.
\tCRITICAL:    A serious error, indicating that the program itself may be unable to continue running.
"""
        self.sprint(message)
        return    
    
    ## custom def for inline activites
    def handle_output(self, out_obj):
        api = pylxca.pylxca_api.lxca_api()
        if out_obj == True:
            self.sprint("Current Log Level is set to " + api.get_log_level())
        else:
            self.sprint("Fail to set Log Level")
        message = """
Possible Log Levels, Please use following values to set desired log level. 

\tDEBUG:        Detailed information, typically of interest only when diagnosing problems.
\tINFO:        Confirmation that things are working as expected.
\tWARNING:    An indication that something unexpected happened, or indicative of some problem in the near future. 
\tERROR:        Due to a more serious problem, the software has not been able to perform some function.
\tCRITICAL:    A serious error, indicating that the program itself may be unable to continue running.
"""
        self.sprint(message)
        return
    
###############################################################################

class ostream(InteractiveCommand):
    """
    Configure output stream or verbose level of command shell 
    
    USAGE:
        ostream -h | --help
        ostream [-l <level>]
    
    OPTIONS:
        -l, --lvl    verbose level

    """
    
    def handle_no_input(self,con_obj = None):
        self.sprint("Current ostream level is set to %s" %(self.shell.ostream.get_lvl()))
        message = """
Possible ostream levels, Please use following values to set desired stdout level. 

\t0:Quite.
\t1:Console.
\t2:File. 
\t3:Console and File.
"""
        self.sprint(message)
        return 
       
    def handle_input(self, dict_handler, con_obj = None):
        lvl = None
        if dict_handler:
            lvl = next((item for item in [dict_handler.get('l') , dict_handler.get('lvl')] if item is not None),None)
            return self.shell.ostream.set_lvl(int(lvl))
        return False
        
    ## custom def for inline activites
    def handle_output(self, out_obj):
        if out_obj == True:
            self.sprint("Current ostream level is set to %s" %(self.shell.ostream.get_lvl()))
        else:
            self.sprint("Fail to set ostream Level")
            message = """
Possible ostream levels, Please use following values to set desired ostream level. 

\t0:Quite.
\t1:Console.
\t2:File. 
\t3:Console and File.
"""
            self.sprint(message)
        return

###############################################################################

class chassis(InteractiveCommand):
    """
    Get Chassis List and Chassis Information
    
    USAGE:
        chassis -h
        chassis [-u <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -u, --uuid    chassis uuid
        -s, --status    chassis manage status (managed/unmanaged)
        -v, --view    view filter name

    """    
    
###############################################################################

class nodes(InteractiveCommand):
    """
    Retrieve nodes List and nodes Information
    
    USAGE:
        nodes -h
        nodes [-u <node UUID>] [-s <managed/unmanaged>] [-c <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -u, --uuid    node uuid
        -s, --status    nodes manage status (managed/unmanaged)
        -c, --chassis    chassis uuid
        -v, --view    view filter name
        -x, --metrics   flag to fetch metrics

    """
    
###############################################################################

class switches(InteractiveCommand):
    """
    Retrieve switches List and switches Information
    
    USAGE:
        switches -h
        switches [-u <switch UUID>] [-c <chassis UUID>] [-v <view filter name>]
        switches  [-u <switch_UUID>] [--ports <port_name>] [--action <action>]
    
    OPTIONS:
        -h            This option displays command usage information
        -u, --uuid    switch uuid
        -c, --chassis    chassis uuid
        --ports        portnames if port is empty lists ports
        --action       enable/disable ports
        -v, --view    view filter name

    """

    def handle_command(self, opts, args):

        # code to handle --ports command without value
        no_args = len(args)
        change = False
        try:
            i = args.index('--ports')
            if i < (no_args - 1):
                next_args = args[i + 1];
                if next_args.startswith("-"):
                    change = True
                else:
                    change = False
            else:
                change = True
        except ValueError:
            change = False
        if change:
            args = [w.replace('--ports', '--ports=') if w == "--ports" else  w for w in args]

        return InteractiveCommand.handle_command(self, opts, args)


###############################################################################

class fans(InteractiveCommand):
    """
    Retrieve fan List and fans Information
    
    USAGE:
        fans -h
        fans [-u <fan UUID>] [-c <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -u, --uuid    fan uuid
        -c, --chassis    chassis uuid
        -v, --view    view filter name

    """
    
###############################################################################

class powersupplies(InteractiveCommand):
    """
    Retrieve Power Supply Information
    
    USAGE:
        powersupplies -h
        powersupplies [-u <power supply UUID>] [-c <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h            This option displays command usage information
        -u, --uuid    power supply uuid
        -c, --chassis    chassis uuid
        -v, --view    view filter name

    """
                
###############################################################################

class fanmuxes(InteractiveCommand):
    """
    Retrieve fan Mux Information

    USAGE:
        fanmuxes -h
        fanmuxes [-u <fan mux UUID>] [-c <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -u, --uuid    fan mux uuid
        -c, --chassis    chassis uuid
        -v, --view    view filter name

    """

###############################################################################

class cmms(InteractiveCommand):
    """
    Get CMM  Information
    
    USAGE:
        cmms -h
        cmms [-u <cmm UUID>] [-c <chassis UUID>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -u, --uuid    cmm uuid
        -c, --chassis    chassis uuid
        -v, --view    view filter name

    """


###############################################################################

class scalablesystem(InteractiveCommand):
    """
    Retrieve Scalable Complex System  Information
    
    USAGE:
        scalablesystem -h
        scalablesystem [-i <scalablesystem id>] [-t <scalablesystem type>] [-s <status>] [-v <view filter name>]
    
    OPTIONS:
        -h        This option displays command usage information
        -i, --id    scalable complex id
        -t, --type    type (flex/rackserver)
        -s, --status    scalable system manage status (managed/unmanaged)
        -v, --view    view filter name

"""

    
###############################################################################

class jobs(InteractiveCommand):
    """
    Retrieve and Manage information about jobs.

    USAGE:
        jobs -h | --help
        jobs [-i <job id>][-u <uuid of endpoint>][-s <jobs state>]
        jobs [-c <cancels the job with specified id>]
        jobs [-d <delete the job with specified id>]
    
    OPTIONS:
        -i, --id=    job id
        -u, --uuid=    uuid of endpoint for which jobs should be retrieved
        -s, --state=    job state to retrieve jobs in specified state.
                The state can be one of the following
                Pending
                Running
                Complete
                Cancelled
                Running_With_Errors
                Cancelled_With_Errors
                Stopped_With_Error
                Interrupted
        -c, --cancel=    cancel job of specified id
        -d, --delete=    delete job of specified id

    """
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Jobs command Failed." )
        elif out_obj == False:
            self.sprint("Jobs command Failed." )
        elif out_obj == True:
            self.sprint("Jobs command succeeded" )
        return
    
###############################################################################

class discover(InteractiveCommand):
    """
    Retrieve a list of devices discovered by SLP discovery.
    
    USAGE:
        discover [-i <IP Address of endpoint>][-j <job ID>]
    
    OPTIONS:
        -i, --ip       One or more IP addresses for each endpoint to be discovered.
        -j, --job      Job ID of discover request
    

    """
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Failed to start Discovery job for selected endpoint " )
        elif type(out_obj) == str:
            self.sprint("Discovery job started, jobId = " + out_obj)
        else:
            self.sprint(out_obj)
        return
    
class manage(InteractiveCommand):
    """
    Manage the endpoint.
    
    USAGE:
        manage  -h | --help
        manage  -i <IP Address of endpoint> -u <user ID to access the endpoint>
                -p <current password to access the endpoint> [-r <recovery password for the endpoint>]
                [-f <Force Manage (True/False)>]
        manage  -i <IP Address of endpoint> -s <stored credential ID> [-r <recovery password for the endpoint>]
                [-f <Force Manage (True/False)>]

        manage  -j <job ID> [-v <view filter name>]

    OPTIONS:
        -i, --ip        One or more IP addresses for each endpoint to be managed.
        -u, --user      user ID to access the endpoint
        -p, --pw        The current password to access the endpoint.
        -r, --rpw       The recovery password to be used for the endpoint.
        -j, --job       Job ID of existing manage request
        -f, --force     Force Manage Boolean flag
        -s  --storedcredential_id  stored credential id
        -v, --view      view filter name
    """
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Failed to start manage job for selected endpoint " )
        else:
            self.sprint("Manage job started, jobId = " + out_obj)
        return
    
    def show_output(self, out_obj, view_filter='default'):
        if out_obj == None:
            self.sprint("Failed to start manage job for selected endpoint " )
        else:
            self.sprint("Manage job started, = " + json.dumps(out_obj))
        return
    
class unmanage(InteractiveCommand):
    """
    Unmanage the endpoint

    USAGE:
        unmanage -h | --help
        unmanage -e <endpoint information> [--force]
        unmanage -j <job ID> [-v <view filter name>]
    
    OPTIONS:
        -e, --ep        one or more endpoints to be unmanaged.
                        This is comma separated list of multiple endpoints, each endpoint should
                        contain endpoint information separated by semicolon.
                        endpoint's IP Address(multiple addresses should be separated by #), UUID of the endpoint and
                        Type of endpoint to be unmanaged ,This can be one of the following values:
                            Chassis
                            ThinkServer
                            Storage
                            Rackswitch
                            Rack-Tower
                            Edge
        -f, --force     Indicates whether to force the unmanagement of an endpoint (True/False)
        -j, --job       Job ID of unmanage request
        -v, --view      View filter name

    """
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Failed to start unmanage job for selected endpoint " )
        else:
            self.sprint("Unmanage job started, jobId = " + out_obj)
        return
    
    def show_output(self, out_obj, view_filter='default'):
        if out_obj == None:
            self.sprint("Failed to start unmanage job for selected endpoint " )
        else:
            self.sprint("Unmanage job started, = " + json.dumps(out_obj))
        return

###############################################################################

class lxcalog(InteractiveCommand):
    """
    Retrieve and Manage information about LXCA Event log.

    USAGE:
        lxcalog [-f < events that apply to the specified filters >]

    """
###############################################################################

class ffdc(InteractiveCommand):
    """
    Retrieve and Manage information about ffdc

    USAGE:
        ffdc -u <UUID of the target endpoint>
    
    OPTIONS:
        -u, --uuid    <UUID of the target endpoint>
    """
    def handle_output(self, out_obj):
        if out_obj == None:
            self.sprint("Failed to start ffdc job for selected endpoint " )
        else:
            self.sprint("FFDC job started, jobId = " + out_obj)
        return

    def show_output(self, out_obj, view_filter='default'):
        if out_obj == None:
            self.sprint("Failed to start ffdc job for selected endpoint " )
        else:
            self.sprint("FFDC job started, jobId = " + out_obj)
        return

###############################################################################

class users(InteractiveCommand):
    """
    Retrieve and Manage information about users.
    
    USAGE:
        users [-i <unique ID of the user to be retrieved>][-v <view filter name>]

    OPTIONS:
        -i, --id    unique ID of the user to be retrieved
        -v, --view    View filter name

    """
###############################################################################

class updatepolicy(InteractiveCommand):
    """
    Retrieve and Manage information about jobs.
    
    USAGE:
        updatepolicy [-v <view filter name>]
        updatepolicy -p <Compliance policy to be assigned to device> -u <UUID of Device> -t <Device Type>
        updatepolicy -j <Job Id of assign policy operation>
        updatepolicy -i <Information type of compliance policy to be retreived>
    
    OPTIONS:
        -p, --policy    Name of the compliance-policy to be assigned to device
        -u, --UUID      UUID of the device to which you want to assign the compliance policy
        -t, --type      Type = The device type. This can be one of the following values.
                            CMM - Chassis Management Module
                            IOSwitch - Flex switch
                            RACKSWITCH - RackSwitch switch
                            STORAGE - Lenovo Storage system
                            SERVER - Compute node or rack server

        -j, --job       Job ID of assign compliance policy operation

        -i, --info      Specifies the type of information to return. This can be one of the following values:
                            FIRMWARE- Returns information about firmware that is applicable to each managed endpoint
                            RESULTS- Returns persisted compare result for servers to which a compliance policy is assigned

        -v, --view      View filter name

    """
###############################################################################
class updaterepo(InteractiveCommand):
    """
    Retrieve and Manage information about jobs.

    USAGE:
        updaterepo -k <Key to return the specified type of update> [-v <view filter name>]
        updaterepo -a <action to take> [-v <view filter name>]
    OPTIONS:
        -k, --key       Returns the specified type of update. This can be one of the following values.
                        supportedMts - Returns a list of supported machine types
                        size - Returns the repository size
                        lastRefreshed - Returns the timestamp of the last repository refresh
                        importDir - Returns the import directory for the repository.
                        publicKeys - Returns the supported signed keys
                        updates - Returns information about all firmware updates
                        updatesByMt - Returns information about firmware updates for the specified machine type
                        updatesByMtByComp - Returns the update component names for the specified machine type
        -a  --action    The action to take. This can be one of the following values.
                        read - Reloads the repository files. The clears the update information in cache and reads the update file again from the repository.
                        refresh - Retrieves information about the latest available firmware updates from the Lenovo Support website,
                            and stores the information to the firmware-updates repository.
                        acquire - Downloads the specified firmware updates from Lenovo Support website, and stores the updates to the firmware-updates repository.
                        delete - Deletes the specified firmware updates from the firmware-updates repository.
                        export.not supported

        -m  --mt        comma separated machine types
        -s  --scope     scope of operation [ all/latest] for refresh and [ payloads ] for acquire
        -f  --fixids    comma separated fixids
        -t  --type      filetype [ all/payloads ]

        -v, --view    View filter name

    """
###############################################################################
class updatecomp(InteractiveCommand):
    """
    Update the firmware of specified component.
        
    USAGE:
        updatecomp [-q <The data to return>] [-v <view filter name>]
        updatecomp  [-m <activate mode>] [-a <action to take>] [-c <information of cmms>] [-w <information of switches>] [-s <information of servers>] [-t <information of storages>]
        updatecomp  -a power [-c <cmms UUID and desired state>] [-w <switches UUID and desired state>]  [-s <servers UUID and desired state>]
    
    OPTIONS:
        -q, --query     The data to return. This can be one of the following values.
                            components - Returns a list of endpoints and components that can be updated.
                            status - Returns the status and progress of firmware updates. This is the default value
        -m, --mode      Indicates when to activate the update. This can be one of the following values.
                            immediate - Uses Immediate Activation mode when applying firmware updates to the selected endpoints.
                            delayed - Uses Delayed Activation mode when applying firmware updates to the selected endpoints.
                            prioritized - Firmware updates on the baseboard management controller are activated immediately

        -a, --action    The action to take. This can be one of the following values.
                            apply - Applies the associated firmware to the submitted components.
                            applyBundle - Applies firmware updates to all components of specified ThinkSystem SR635 and SR655 devices according to the assigned firmware-compliance policy using a bundled image that contain all applicable firmware packages.
                            power - Perform power action on selected endpoint.
                            cancelApply - Cancels the firmware update request to the selected components.
        -c, --cmm       cmms information
        -w, --switch    switch information
        -s, --server    servers information
        -t, --storage   storages information
    
            For action = apply / applyBundle / cancelApply, Device information should contain following data separated by comma
                UUID - UUID of the device
                Fixid - Firmware-update ID of the target package to be applied to the component. If not provided assigned policy would be used.
                Component - Component name

            For action = power, Device information should contain following data separated by comma
                UUID - UUID of the device
                powerState - One of the power state values. Possible values per device type are
                    Server: powerOn, powerOff, powerCycleSoft, powerCycleSoftGraceful, powerOffHardGraceful
                    Switch: powerOn, powerOff, powerCycleSoft
                    CMM: reset
                    Storage:powerOff,powerCycleSoft

        -v, --view      View filter name

    """

    
###############################################################################
class configtargets(InteractiveCommand):
    """
    Retrieve and Manage information of configuration targes.
    
    USAGE:
        configtargets [-i <ID of specific profile or pattern>] [-v <view filter name>]
    
    OPTIONS:
        -i, --id    The unique ID that was assigned when the server pattern was created
        -v, --view    View filter name
    
    
    """
###############################################################################


class configpatterns(InteractiveCommand):
    """
    Retrieve and Manage information about config patterns.
    
    USAGE:
        configpatterns [-i <ID of specific pattern>] [-r <when to activate the configurations>] [-e <Comma separated list of one or more UUIDs for the target servers>] [-t <type of the target server>] [-s True] [-v <view filter name>]
    
    OPTIONS:
        -i, --id    The unique ID that was assigned when the server pattern was created
        -n, --name  Name of pattern
        -e, --endpoint    Comma separated list of one or more UUIDs for the target servers,If a target is an empty bay,
                specify the location ID; otherwise, specify the server UUID
        -r, --restart    When to activate the configurations. This can be one of the following values:
                defer - Activate IMM settings but do not restart the server.
                immediate - Activate all settings and restart the server immediately.
                pending - Manually activate the server profile and restart the server.
        -t, --type    Type of the server, It can be one of the following
                Flex  - for empty bay having endpoint other than UUID
                Node
                Rack
                Tower
        -s, --status  return config status
        -v, --view    View filter name

    """

    def handle_output(self, out_obj):
        if 'compliance' in out_obj:
            self.sprint(out_obj['message'])
        else:
            self.sprint(out_obj)
        return
    
    def show_output(self, out_obj, view_filter='default'):
        if 'compliance' in out_obj:
            self.sprint(out_obj['message'])
        else:
            self.sprint(out_obj)
        return

###############################################################################

class configprofiles(InteractiveCommand):
    """
    Retrieve and Manage information of config profiles.

    USAGE:
        configprofiles [-i <ID of specific profile>] [-v <view filter name>]
        configprofiles -i <ID of specific profile> -n <New profile name>
        configprofiles -i <ID of specific profile> -e <UUID or location ID>  -r <restart option> [-v <view filter name>]
        configprofiles -i <ID of specific profile>  -d <boolean>   [-v <view filter name>]
        configprofiles -i <ID of specific profile>  -u <boolean> -p <boolean> --resetimm <boolean> -f <boolean> [-v <view filter name>]

    OPTIONS:
        -i, --id    The unique ID that was assigned when the server pattern was created
               id          The unique ID that was assigned when the server profile was created
        -n, --name        profile name
        -e,  --endpoint    endpoint  UUID of the server or location id for flex system
        -r,  --restart     restart server to activate profile ( immediate / defer )
        -d,  --delete      True for delete id
        -u   --unassign    unassign specified id
        -p   --powerdown   powerdown server [true/false]
             --resetimm    reset IMM [true/false]
        -f   --force       force profile deactivation [true/false]
        -v, --view    View filter name

    """

    def handle_output(self, out_obj):
        if 'compliance' in out_obj:
            self.sprint(out_obj['message'])
        else:
            self.sprint(out_obj)
        return

    def show_output(self, out_obj, view_filter='default'):
        if 'compliance' in out_obj:
            self.sprint(out_obj['message'])
        else:
            self.sprint(out_obj)
        return

###############################################################################

class manifests(InteractiveCommand):
    """
    Send solution manifest to and retreive manifests from Lenovo XClarity Administrator.
    """

###############################################################################

class tasks(InteractiveCommand):
    """
    Retrieve tasks List and tasks Information

    USAGE:
        tasks -h
        tasks [-u <JOB UUID>]  [-v <view filter name>]

    OPTIONS:
    -h            This option displays command usage information
        -u, --uuid    Job uuids
        -v, --view    view filter name

    """

###############################################################################
class resourcegroups(InteractiveCommand):
    """
    create Group of Resources
    """

###############################################################################



class osimages(InteractiveCommand):
    """
    OSImages/Deployment on LXCA
    """
    
    def handle_output(self, out_obj):
        if (
            'result' in out_obj and 
            out_obj['result'] == 'failed' and 
            out_obj['messages'][0]['id'] == 'FQXHMFC0295M'
        ):
            self.sprint(out_obj['messages'][0]['text'])
            self.sprint(out_obj['messages'][0]['recovery']['text'])
        else:
            self.sprint(out_obj)
        return
    
    def show_output(self, out_obj, view_filter='default'):
        if (
            'result' in out_obj and 
            out_obj['result'] == 'failed' and 
            out_obj['messages'][0]['id'] == 'FQXHMFC0295M'
        ):
            self.sprint(out_obj['messages'][0]['text'])
            self.sprint(out_obj['messages'][0]['recovery']['text'])
        else:
            self.sprint(out_obj)
        return


###############################################################################



class managementserver(InteractiveCommand):
    """
    managementserver update on LXCA
    """

###############################################################################



class rules(InteractiveCommand):
    """
    complaince rules get and update on LXCA
    """

###############################################################################



class compositeResults(InteractiveCommand):
    """
    complaince compositeresult get and update on LXCA
    """

class storedcredentials(InteractiveCommand):
    """
        storedcredentials get and update on LXCA

    USAGE:
        storedcredentials -h
        storedcredentials [-u <admin>] [-p <password>] [-d <description>] [-v <view filter name>]
        storedcredentials [-u <Stored Credential Id>] [-u <admin>] [-p <password>] [-d <description>] [-v <view filter name>]

    OPTIONS:
        -h        This option displays command usage information
        -i  --id      stored credential id
        -u, --user_name     user name
        -p, --password    password
        -d, --description detail about user
            --delete_id    credential id to be deleted
        -v, --view    view filter name

    """

###############################################################################

class license(InteractiveCommand):
    """
    Retrieve information about warnings regarding non-compliance of installed licenses.
    """
    def handle_output(self, out_obj):
        self.sprint(out_obj)
        return
    
    def show_output(self, out_obj, view_filter='default'):
        out_obj['response'][1]['notifications'] = out_obj['response'][1]['notifications'].pop()
        out_obj['response'] = [self.merge_two_dicts(out_obj['response'][0], out_obj['response'][1])]
        super().show_output(out_obj, view_filter)
        return

    def merge_two_dicts(self, x, y):
        z = x.copy()
        z.update(y)
        return z


###############################################################################



# Package init file

'''LXCA Python Client API'''

#setting CMD_PATH environment variable to be used in submodules for loading data files
import os
pylxca_cmd_path = os.path.dirname(__file__)
os.environ['PYLXCA_CMD_PATH'] = pylxca_cmd_path

# All submodules of this package are imported; so clients need to import just this package.
from pylxca.pylxca_cmd import lxca_ishell
from pylxca.pylxca_cmd import lxca_icommands
from pylxca.pylxca_api import *
import pylxca.pylxca_cmd


# Version of the pylxca package

__version__ = '4.0.0'


# There are submodules, but clients shouldn't need to know about them.
# Importing just this module is enough.

# These are explicitly safe for 'import *'
from pylxca.pylxca_api import *
from pylxca.pylxca_cmd import *
from pylxca.pylxca_cmd.lxca_pyshell import *

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

#Configure Logger
import logging.config
pylxca.pylxca_api.lxca_rest().set_log_config()
pyshell()



'''
@since: 4 Sep 2015
@author: Prashant Bhosale <pbhosale@lenovo.com>, Girish Kumar <gkumar1@lenovo.com>
@license: Lenovo License
@copyright: Copyright 2016, Lenovo
@organization: Lenovo
@summary: This module is for creating a connection session object for given xHMC
'''

import logging, os, json, pprint, requests
import logging.config
from requests_toolbelt import (MultipartEncoder,
                               MultipartEncoderMonitor)
from requests.exceptions import HTTPError
import json, re
import socket
import time

import urllib


try:
    logging.captureWarnings(True)
except:
    pass

logger_conf_file = "lxca_logger.conf"
pylxca_logger = os.path.join(os.getenv('PYLXCA_API_PATH'), logger_conf_file)

logger = logging.getLogger(__name__)
REST_TIMEOUT = 60


def callback(encoder):
    # uncomment it to debug, spit lot of data in log file for big upload
    #logger.debug("Callback called with data length %d" % (encoder.bytes_read))
    pass


class lxca_rest(object):
    '''
    classdocs
    '''
    def get_chassis(self,url, session, uuid, status ):
        url = url + '/chassis'

        if uuid:
            url = url + "/" + uuid

        if status:
            url = url + "?status=" + status
        else:
            url = url + "?status=managed"

        try:
            resp = session.get(url,verify=False,timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp    
        
    def get_metrics(self, url, session, uuid ):
        url = url + '/nodes/metrics'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        
        return resp

    def get_nodes(self,url, session, uuid, status):
        url = url + '/nodes'

        if uuid:
            url = url + '/' + uuid

        if status:
            url = url + "?status=" + status

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def set_nodes(self, url, session, uuid, modify):
        url = url + '/nodes'
        job = None

        if uuid:
            url = url + '/' + uuid + '?synchronous=false'

        try:
            payload = dict()
            payload = modify

            resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                if "location" in resp.headers._store:
                    job = resp.headers._store["location"][-1].split("/")[-1]

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re

        return job


    def get_switches(self,url, session, uuid):
        url = url + '/switches'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_switches_port(self,url, session, uuid, list_port):
        url = url + '/switches'

        if uuid:
            url = url + '/' + uuid
        else:
            raise Exception("Invalid argument uuid is required")

        if list_port:
            url = url + '/ports'

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def put_switches_port(self, url, session, uuid, port_name, action):
        url = url + '/switches'

        if uuid:
            url = url + '/' + uuid
        else:
            raise Exception("Invalid argument uuid is required")

        if port_name:
            url = url + '/ports'
        else:
            raise Exception("Invalid argument port name is required")

        if action not in ['enable', 'disable']:
            raise Exception("Invalid argument action [enable/disable] is required %s" %action)

        payload = {
                       "action":"enable",
                       "ports":[]
                }
        uuid_list = port_name.split(",")
        payload["action"] = action
        payload["ports"] = uuid_list

        try:
            resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_fan(self,url, session, uuid):
        url = url + '/fans'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_powersupply(self,url, session, uuid):
        url = url + '/powerSupplies'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_fanmux(self,url, session, uuid):
        url = url + '/fanMuxes'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_cmm(self,url, session, uuid):
        url = url + '/cmms'

        if uuid:
            url = url + '/' + uuid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re
        return resp

    def get_scalablesystem(self,url, session, complexid,complextype):
        url = url + '/scalableComplex'

        if complexid:
            url = url + '/' + complexid

        if complextype:
            if complextype == "flex" or complextype == "rackserver":
                url = url + "?complexType=" + complextype
            else:
                raise Exception("Invalid argument 'complexType': %s" %complextype)

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def set_log_config(self):
        logging.config.fileConfig(pylxca_logger)
        return

    def get_log_level(self):
        logger.debug("Current Log Level is: " + str(logger.getEffectiveLevel()))
        return logging.getLevelName(logger.getEffectiveLevel())

    def set_log_level(self,log_value):
        logger.setLevel(log_value)
        for handler in logger.handlers:
            handler.setLevel(log_value)
        logger.debug("Current Log Level is: " + str(logger.getEffectiveLevel()))
        return

    def do_discovery(self,url, session, ip_addr,jobid):
        try:
            if ip_addr:
                url = url + '/discoverRequest'
                payload = [{"ipAddresses":ip_addr.split(",")}]
                resp = session.post(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    if "location" in resp.headers._store:
                        job = resp.headers._store["location"][-1].split("/")[-1]
                        return job
                    else:
                        return None
            elif jobid:
                url = url + '/discoverRequest/jobs/' + str(jobid)
                resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
            else:
                url = url + '/discovery'
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re
        return resp.json()

    def do_manage(self,url, session, ip_addr,user,pw,rpw,force,jobid, storedcredential_id):
        try:
            orig_url = url
            #All input arguments ip_add, user, pw, rpw and mp are mandatory
            # if ip_addr and user and pw and mp:
            if ip_addr and \
                    (( user and pw) or (storedcredential_id)):
                url = url + '/manageRequest'

                payload = list()
                param_dict = dict()

                param_dict["ipAddresses"]=ip_addr.split(",")
                param_dict["username"] = user
                param_dict["password"] = pw
                if rpw:param_dict["recoveryPassword"] = rpw

                # do auto discovery
                disc_job_id = self.do_discovery(url.rsplit('/',1)[0], session, ip_addr,None)
                disc_progress = 0
                
                if disc_job_id:
                    #TODO Check if dicovery succeed, if not throw Exception
                    while disc_progress < 100:
                        time.sleep(2) # delays for 5 seconds to allow discovery to complete
                        disc_job_resp = self.do_discovery(url.rsplit('/',1)[0], session, None,disc_job_id)
                        disc_resp_py_obj = disc_job_resp
                        disc_progress = disc_resp_py_obj['progress']
                
                discovered_endpoint = False
                for key in list(disc_resp_py_obj.keys()):
                    if isinstance(disc_resp_py_obj[key],list) and disc_resp_py_obj[key] != []: 
                        discovered_endpoint = True
                        
                        #Fetch Management Port value from Response
                        param_dict["managementPorts"] = disc_resp_py_obj[key][0]["managementPorts"]
                        
                        #Fetch Type value from Response
                        param_dict["type"] = disc_resp_py_obj[key][0]["type"]

                        # Fetch machine type from Response
                        param_dict["machineType"] = disc_resp_py_obj[key][0]["machineType"]

                        if param_dict["type"] == "Rack-Tower Server" or param_dict["type"] == "Edge Server":
                        # Fetch ManagementProcessor value from Response
                            param_dict["managementProcessor"] = disc_resp_py_obj[key][0]["managementProcessor"]
                            param_dict['server-type'] = param_dict['type']

                        #Fetch UUID value from  Response
                        param_dict["uuid"] = disc_resp_py_obj[key][0]["uuid"]
                        
                        disc_ip_addr = disc_resp_py_obj[key][0]["ipAddresses"][0]
                        
                        param_dict["ipAddresses"] = [disc_ip_addr]
                        
                        if param_dict["type"] == "Rackswitch":
                            param_dict["os"] = disc_resp_py_obj[key][0]["os"]
                
                if not discovered_endpoint:
                    logger.debug("Failed to discover given endpoint  %s" %param_dict["ipAddresses"])
                    raise Exception("Failed to discover given endpoint  %s" %param_dict["ipAddresses"])
                
                if force:
                    if isinstance(force, bool):
                        param_dict["forceManage"] = force
                    else:
                        if force.lower() == "true":
                            param_dict["forceManage"] = True
                        else:
                            param_dict["forceManage"] = False

                security_Descriptor = { }
                if storedcredential_id:
                    security_Descriptor['managedAuthEnabled'] = False
                    security_Descriptor['managedAuthSupported'] = False
                    cred = self.get_storedcredentials(orig_url, session, storedcredential_id)
                    cred_resp = cred.json()
                    storedCredentials = {}
                    storedCredentials['id'] = storedcredential_id
                    storedCredentials['userName'] = cred_resp['response']['userName']
                    storedCredentials['description'] = cred_resp['response']['description']
                    security_Descriptor['storedCredentials'] = storedCredentials
                else:
                    security_Descriptor['managedAuthEnabled'] = True
                    security_Descriptor['managedAuthSupported'] = False

                param_dict['securityDescriptor'] = security_Descriptor
                payload = [param_dict]

                resp = session.post(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()

                manage_resp = {}
                manage_resp['manageRequestJob'] = None

                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    if "location" in resp.headers._store:
                        job = resp.headers._store["location"][-1].split("/")[-1]
                        manage_resp['manageRequestJob'] = job

                manage_resp['manageRequestResponse'] = resp.json()

                return manage_resp
                    
            elif jobid:
                url = url + '/manageRequest/jobs/' + str(jobid)
                resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp.json()
            else:
                logger.error("Invalid execution of manage REST API")
                raise Exception("Invalid execution of manage REST API")

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re
        return resp

    def do_unmanage(self,url, session, endpoints,force,jobid):

        endpoints_list = list()
        param_dict = dict()

        try:
            if endpoints:
                url = url + '/unmanageRequest'
                for each_ep in endpoints.split(","):
                    ip_addr = None
                    each_ep_dict = dict()

                    ep_data = each_ep.split(";")
                    ip_addr = ep_data[0]
                    uuid = ep_data[1]
                    type = ep_data[2]
                    #Fetch type value from input
                    type_list = ["Chassis","Rackswitch","ThinkServer","Storage","Rack-Tower","Edge"]
                    if type not in type_list:
                        raise Exception("Invalid Type Specified")
                    if type == "ThinkServer": type = "Lenovo ThinkServer"
                    elif type == "Storage": type = "Lenovo Storage"
                    elif type == "Rack-Tower": type = "Rack-Tower Server"
                    elif type == "Edge": type = "Edge Server"
                    each_ep_dict = {"ipAddresses":ip_addr.split("#"),"type":type,"uuid":uuid}
                    endpoints_list.append(each_ep_dict)
                param_dict["endpoints"] = endpoints_list

                if force:
                    if isinstance(force, bool):
                        param_dict["forceUnmanage"] = force
                    else:
                        if force.lower() == "true":
                            param_dict["forceUnmanage"] = True
                        else:
                            param_dict["forceUnmanage"] = False

                payload = param_dict
                resp = session.post(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()

                unmanage_resp = {}
                unmanage_resp['unmanageRequestJob'] = None

                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    if "location" in resp.headers._store:
                        job = resp.headers._store["location"][-1].split("/")[-1]
                        unmanage_resp['unmanageRequestJob'] = job

                unmanage_resp['unmanageRequestResponse'] = resp.json()

                return unmanage_resp

            elif jobid:
                url = url + '/unmanageRequest/jobs/' + str(jobid)
                resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp.json()
            else:
                logger.error("Invalid execution of unmanage REST API")
                raise Exception("Invalid execution of unmanage REST API")

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

        return resp

    def get_jobs(self,url, session,jobid,uuid,state,canceljobid,deletejobid):
        url = url + '/jobs'
        try:
            if jobid:
                url = url + '/' + jobid

                if state:
                    if state == "Pending " or state == "Running" \
                    or state == "Complete" or state == "Cancelled" \
                    or state == "Running_With_Errors" or state == "Cancelled_With_Errors" \
                    or state == "Stopped_With_Error" or state == "Interrupted":
                        url = url + '?state=' + state
                        if uuid:
                            url = url + ',uuid=' + uuid
                    else:
                        raise Exception("Invalid argument 'state': %s" %state)
                if state == None and uuid:
                    url = url + '?uuid=' + uuid

                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
            elif canceljobid:
                url = url + '/' + canceljobid
                payload = {"cancelRequest":"true"}
                resp = session.put(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    return True
                resp.raise_for_status()
            elif deletejobid:
                url = url + '/' + deletejobid
                resp = session.delete(url,verify=False, timeout=REST_TIMEOUT)
                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    return True
                resp.raise_for_status()
            else:
                if state:
                    if state == "Pending" or state == "Running" \
                    or state == "Complete" or state == "Cancelled" \
                    or state == "Running_With_Errors" or state == "Cancelled_With_Errors" \
                    or state == "Stopped_With_Error" or state == "Interrupted":
                        url = url + '?state=' + state
                        if uuid:
                            url = url + ',uuid=' + uuid
                    else:
                        raise Exception("Invalid argument 'state': %s" %state)
                if state == None and uuid:
                    url = url + '?uuid=' + uuid

                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_users(self,url, session, userid):
        url = url + '/userAccounts'

        if userid:
            url = url + '/' + userid

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re
        return resp

    def get_lxcalog(self,url, session, filter):
        url = url + '/events'

        if filter:
            url = url + '?filterWith=' + filter

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re
        return resp

    def get_ffdc(self,url, session, uuid):
        url = url + '/ffdc/endpoint'
        try:
            if uuid:
                url = url + '/' + uuid
                resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                if resp.status_code == requests.codes['ok'] or resp.status_code == requests.codes['created'] or resp.status_code == requests.codes['accepted']:
                    job_info = resp.json()
                    if "jobURL" in job_info:
                        job = job_info["jobURL"].split("/")[-1]
                        return job
                    else:
                        return resp.json()
            else:
                logger.error("Invalid execution of ffdc REST API mandatory parameter uuid is missing")
                raise Exception("Invalid execution of ffdc REST API mandatory parameter uuid is missing")

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

    def get_updatepolicy(self, url, session, info,  jobid, uuid):
        url = url + '/compliancePolicies'
        try:
            if info in ["FIRMWARE", "RESULTS", "NAMELIST"]:
                if info == "FIRMWARE":
                    url = url + "/applicableFirmware"
                elif info == "RESULTS":
                    url = url + "/persistedResult"
                elif info == "NAMELIST":
                    url = url + "/nameList"
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp
            elif jobid:
                url = url + "/compareResult"
                payload = dict()
                payload["jobid"] = jobid
                payload["uuid"] = uuid
                resp = session.get(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp

            url = url + "?basic_full=full"
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

    def post_updatepolicy(self, url, session, policy, type, uuid):
        url = url + '/compliancePolicies/compareResult'
        try:

            if not policy or not type or not uuid:
                raise Exception("Invalid argument key")

            payload = dict()

            policy_dict = dict()
            policy_dict["policyName"] = policy

            # do auto discovery
            # disc_job_resp = self.do_discovery(url.rsplit('/', 2)[0], session, None, None)
            # disc_resp_py_obj = json.loads(disc_job_resp.text)
            #
            # for key in disc_resp_py_obj.keys():
            #     if isinstance(disc_resp_py_obj[key], list) and disc_resp_py_obj[key] != []:
            #         # Fetch UUID value from  Response
            #         for item in disc_resp_py_obj[key]:
            #             if uuid == item["uuid"]:
            #                 # Fetch Type value from Response
            #                 policy_dict["type"] = disc_resp_py_obj[key][0]["type"]
            #
            # logger.debug("Type for %s = %s" %(uuid, policy_dict["type"]))

            policy_dict["uuid"] = uuid
            policy_dict["type"] = type
            compliance_list = []
            compliance_list.append(policy_dict)
            payload['compliance'] = compliance_list

            resp = session.post(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re


    def get_updaterepo(self, url, session, key, mt, scope):
        url = url + '/updateRepositories/firmware'
        try:
            if key:
                url= url + "?key=" + key

            if mt:
                url = url + "&mt=" + mt

            if scope:
                if scope.lower() in ["all", "latest"]:
                    if key == "updates" or key == "updatesByMt":
                        url = url + "&with=" + scope.lower()
                    else:
                        raise Exception("Invalid argument combination of key and scope")
                else:
                    raise Exception("Invalid argument scope: " + scope)

            resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

    def put_updaterepo(self, url, session, action , fixids, mt, type, scope):
        url = url + '/updateRepositories/firmware'
        try:
            url= url + "?action=" + action

            if type:
                if type.lower() in ["all", "payloads"]:
                    if action == "delete" or action == "export":
                        url = url + "&filetypes=" + type.lower()
                    else:
                        raise Exception("Invalid argument combination of action and type")
                else:
                    raise Exception("Invalid argument type:" + type)

            if scope:
                if scope.lower() in ["all", "latest", "payloads"]:
                    if action == "refresh" and scope.lower() in ["all", "latest"]:
                        url = url + "&with=" + scope.lower()
                    elif action == "acquire" and scope.lower() in ["payloads"]:
                        url = url + "&with=" + scope.lower()
                    else:
                        raise Exception("Invalid argument combination of action and scope")
                else:
                    raise Exception("Invalid argument scope:" + scope)

            payload = dict()
            if action == "delete":
                if fixids:
                    fixids_list = fixids.split(",")
                    payload['fixids'] = fixids_list
                else:
                    raise Exception("Invalid argument fixids is required for delete")

            if action == "acquire":
                if fixids:
                    fixids_list = fixids.split(",")
                    payload['fixids'] = fixids_list

            if action == "acquire" or action == "refresh":
                if mt:
                    mt_list = mt.split(",")
                    payload['mt'] = mt_list
                else:
                    raise Exception("Invalid argument mt is required for action acquire and refresh")

            if action == "refresh":
                payload['os'] = ""
                payload['type'] = "catalog"

            if action == "acquire":
                payload['type'] = "latest"

            resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

    def get_managementserver(self, url, session, key, fixids, type):
        url = url + '/managementServer/updates'
        try:
            if fixids:
                url = url + "/" + fixids

                if key:
                    if key not in ['all']:
                        url = url + "?key=" + key
                elif type:
                    url = url + "?filetype=" + type

            else:
                if key:
                    if key not in ['all']:
                        url = url + "?key=" + key

            resp = session.get(url,verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re

    def set_managementserver(self, url, session, action, files, jobid, fixids):
        url = url + '/managementServer/updates'
        try:
            if not action  == None \
                    and action == "apply":
                # implement PUT
                url = url + "?action=apply"
                payload = {}

                if fixids:
                    payload['fixids'] = [fixids]
                else:
                    raise Exception("Invalid argument apply requires fixids")
                resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                return resp
            # Creations of Import job POST
            if not action == None and action == "import":
                file_list = files.strip().split(",")
                file_type_dict = {'.txt': 'text/plain',
                                  '.xml': 'text/xml',
                                  '.chg': 'application/octet-stream',
                                  '.tgz': 'application/x-compressed'}

                if jobid == None:
                    url = url + "?action=import"
                    #payload = {"files":[{"index":0,"name":"lnvgy_sw_lxca_thinksystemrepo1-1.3.2_anyos_noarch.xml","size":7329,"type":"text/xml"}]}
                    payload_files = [{
                                         'index': index,
                                         'name': os.path.basename(file),
                                         'size': os.path.getsize(file),
                                         'type': file_type_dict[os.path.splitext(os.path.basename(file))[-1]]
                                     } for index, file in enumerate(file_list)]
                    payload = {'files' : payload_files}
                    resp = session.post(url, data=json.dumps(payload), verify=False, timeout= 2 * REST_TIMEOUT)
                    return resp

                else :
                    url = url.replace('/managementServer/updates', '/files/managementServer/updates')
                    url = url + "?action=import&jobid=" + jobid
                    m = MultipartEncoder(
                        fields=[('uploadedfile[]', (os.path.basename(file),
                                                    open(file, 'rb'),
                                                    file_type_dict[os.path.splitext(os.path.basename(file))[-1]])
                                 ) for file in file_list]
                    )
                    monitor = MultipartEncoderMonitor(m, callback)
                    resp = session.post(url, data = monitor, headers={'Content-Type': monitor.content_type}, verify=False, timeout=100 * REST_TIMEOUT)
                    return resp

            if not action == None \
                    and action == "acquire":
                url = url + "?action=acquire"
                payload = {}
                if fixids:
                    fixids_list = fixids.split(",")
                    payload['fixids'] = fixids_list
                else:
                    raise Exception("Invalid argument key action: acquire requires fixids ")
                resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                return resp

            if not action == None \
                    and action == "refresh":
                url = url + "?action=refresh"
                payload = {}
                payload['mts'] = ['lxca']
                resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                return resp

            if not action == None \
                    and action == "delete":
                # implement delete
                if fixids:
                    url = url + "/" + fixids
                else:
                    raise Exception("Invalid argument key action: delete requires fixids ")

                resp = session.delete(url,  verify=False, timeout=REST_TIMEOUT)
                return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re


    def do_updatecomp(self, url, session, query, mode, action, server, switch, storage, cmm, dev_list):

        if dev_list:
            resp = self.do_updatecomp_all(url, session, action, mode,dev_list)
            return resp

        serverlist = list()
        storagelist = list()
        cmmlist = list()
        switchlist = list()

        url = url + '/updatableComponents'
        try:

            # For Query Action
            if mode == None and action == None and server == None and switch == None and storage == None and cmm == None:
                if query and query.lower() == 'components':
                    url = url + "?action=getComponents"
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp

            # For apply and cancelApply Action
            if action == "apply" or action == "applyBundle" or action == "cancelApply" :
                
                url= url + "?action=" + action
                
                if not mode  == None and mode == "immediate" or mode == "delayed" or mode == "prioritized":
                    url= url + "&activationMode=" + mode
                else:
                    if action != "applyBundle":
                        raise Exception("Invalid argument mode")

                if server:
                    if len(server.split(","))==3:
                        server_data = server.split(",")
                        serverlist = [{"UUID": server_data[0],"Components": [{"Fixid": server_data[1],"Component": server_data[2]}]}]
                    elif len(server.split(","))==2:
                        server_data = server.split(",")
                        serverlist = [{"UUID": server_data[0],"Components": [{"Component": server_data[1]}]}]
                    elif len(server.split(","))==1:
                        uuid = server
                        serverlist = [{"UUID": uuid}]

                if switch:
                    if len(switch.split(","))==3:
                        switch_data = switch.split(",")
                        switchlist = [{"UUID": switch_data[0],"Components": [{"Fixid": switch_data[1],"Component": switch_data[2]}]}]
                    elif len(switch.split(","))==2:
                        switch_data = switch.split(",")
                        switchlist = [{"UUID": switch_data[0],"Components": [{"Component": switch_data[1]}]}]

                if storage:
                    if len(storage.split(","))==3:
                        storage_data = storage.split(",")
                        storagelist = [{"UUID": storage_data[0],"Components": [{"Fixid": storage_data[1],"Component": storage_data[2]}]}]
                    elif len(storage.split(","))==2:
                        storage_data = storage.split(",")
                        storagelist = [{"UUID": storage_data[0],"Components": [{"Component": storage_data[1]}]}]

                if cmm:
                    if len(cmm.split(","))==3:
                        cmm_data = cmm.split(",")
                        cmmlist = [{"UUID": cmm_data[0],"Components": [{"Fixid": cmm_data[1],"Component": cmm_data[2]}]}]
                    elif len(cmm.split(","))==2:
                        cmm_data = cmm.split(",")
                        cmmlist = [{"UUID": cmm_data[0],"Components": [{"Component": cmm_data[1]}]}]

            elif action == "power" : 
                url= url + "?action=" + "powerState"
                
                if server and len(server.split(","))==2:
                    server_data = server.split(",")
                    serverlist = [{"UUID": server_data[0],"PowerState": server_data[1]}]
    
                if switch and len(switch.split(","))==2:
                    switch_data = switch.split(",")
                    switchlist = [{"UUID": switch_data[0],"PowerState": switch_data[1]}]
                
                if storage and len(storage.split(","))==2:
                    storage_data = storage.split(",")
                    storagelist = [{"UUID": storage_data[0],"PowerState": storage_data[1]}]
        
                if cmm and len(cmm.split(","))==2:
                    cmm_data = cmm.split(",")
                    cmmlist = [{"UUID": cmm_data[0],"PowerState": cmm_data[1]}]
                
            param_dict = dict()
            if serverlist:param_dict["ServerList"] = serverlist
            if storagelist:param_dict["StorageList"] = storagelist
            if cmmlist:param_dict["CMMList"] = cmmlist
            if switchlist:param_dict["SwitchList"] = switchlist

            payload = dict()
            payload["DeviceList"] = [param_dict]
            logger.debug("Update Firmware payload: " + str(payload))
            resp = session.put(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s",re)
            raise re


    def do_updatecomp_all(self, url, session, action, mode, dev_list):
        try:
            url = url + '/updatableComponents'
            if action == "apply" or action == "applyBundle" or action == "cancelApply":
                url = url + "?action=" + action

                if not mode == None and mode == "immediate" or mode == "delayed" or mode == "prioritized":
                    url = url + "&activationMode=" + mode
                else:
                    if action != "applyBundle":
                        raise Exception("Invalid argument mode")
            else:
                raise Exception("Invalid argument action")

            #payload = dict()
            #payload["DeviceList"] = dev_list
            payload_data = json.dumps(dev_list)
            logger.debug("Update Firmware payload: " + str(payload_data))

            resp = session.put(url, data=payload_data, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
            return resp

        except HTTPError as re:
            logger.error("Exception occured: %s", re)
            raise re

    def get_configprofiles(self,url, session, profileid):
        url = url + '/profiles'
        
        if profileid:
            url = url + '/' + str(profileid)

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re
    
        return resp

    def put_configprofiles(self, url, session, profileid, profilename):
        url = url + '/profiles'

        if profileid:
            url = url + '/' + str(profileid)
        else:
            raise Exception("Invalid argument ")

        try:
            payload = dict()
            payload['profileName'] = profilename
            resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def post_configprofiles(self, url, session, profileid, endpoint, restart):
        url = url + '/profiles'

        if profileid:
            url = url + '/' + str(profileid)
        else:
            raise Exception("Invalid argument ")

        try:
            payload = dict()
            if restart and endpoint:
                payload['restart'] = restart
                payload['uuid'] = endpoint
            else:
                raise Exception("Invalid argument, restart and endpoint ")

            resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def delete_configprofiles(self, url, session, profileid):
        url = url + '/profiles'

        if profileid:
            url = url + '/' + str(profileid)

        try:
            resp = session.delete(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def unassign_configprofiles(self, url, session, profileid, powerdown, resetimm, resetswitch, force):
        url = url + '/profiles/unassign'

        if profileid:
            url = url + '/' + str(profileid)
        else:
            raise Exception("Invalid argument, profile id is required for unassign ")

        payload = dict()
        if powerdown:
            if powerdown.lower() == "true":
                payload['powerDownITE'] = True
            else:
                payload['powerDownITE'] = False
        if resetimm:
            if resetimm.lower() == "true":
                payload['resetIMM'] = True
            else:
                payload['resetIMM'] = False
        if resetswitch:
            if resetimm.lower() == "true":
                payload['resetSwitch'] = True
            else:
                payload['resetSwitch'] = False


        if force:
            if isinstance(force, bool):
                payload["force"] = force
            else:
                if force:
                    if force.lower() == "true":
                        payload['force'] = True
                    else:
                        payload['force'] = False

        try:
            resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def _validate_uuids(self, uuids):
        uuids_list = uuids.split(",")
        for uuid in uuids_list:
            if len(uuid) < 16:
                raise Exception("Invalid Arguments, uuid : %s" %uuid)



    def do_configpatterns(self, url, session, id, includeSettings, endpoint, restart, etype, pattern_update_dict):
        '''
        input_dict = {}
        input_dict['id'] = id
        input_dict['includeSettings'] = includeSettings
        input_dict['endpoint'] = endpoint
        input_dict['restart'] = restart
        input_dict['type'] = etype
        input_dict['pattern_update_dict'] = pattern_update_dict

        valid_arg_combination = [[], ['id'], ['id', 'includeSettings'],
                                 ['id', 'endpoint', 'restart', 'type'],
                                 ['pattern_update_dict']]

        valid, combination = _validate_combination(input_dict, valid_arg_combination)
        if not valid:
            raise Exception("Invalid Missing Arguments %s" %str(combination))

        '''
        resp = None
        url = url + '/patterns'

        if endpoint:
            self._validate_uuids(endpoint)

        if id != None:
            if len(id) == 0:
                raise Exception("Invalid Argument, id ")
            elif not id.isdigit():
                raise Exception("Invalid Argument, id is not Numeric ")

            url = url + '/' + id
            if includeSettings:
                url = url + '/includeSettings'

        try:
            if endpoint and restart and etype:
                param_dict = dict()
                
                if etype.lower() == 'node' or etype.lower() == 'rack' or etype.lower() == 'tower':
                    param_dict['uuid'] = [endpoint] 
                elif  etype.lower() == 'flex':
                    param_dict['endpointIds'] = [endpoint]
                
                param_dict['restart'] = restart
                
                payload = param_dict
                resp = session.post(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            elif pattern_update_dict:
                payload = dict()
                payload = pattern_update_dict
                resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            else:
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                
            resp.raise_for_status()
        except HTTPError as re:
            raise re
    
        return resp


    def get_configstatus(self, url, session, endpoint):
        resp = None
        url = url + '/config/server'

        if endpoint:
            url = url + '/' + endpoint + '/status'
        else:
            raise Exception("Invalid argument endpoint uuid is required for config status")

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def get_configtargets(self,url, session, targetid):
        url = url + '/config/target'
        
        if targetid:
            url = url + '/' + targetid
        else:
            raise Exception("Invalid argument ID")
        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re
    
        return resp  

    def get_tasks(self,url, session, job_uid, includeChildren):
        url = url + '/tasks'

        if job_uid:
            url = url + '/' + job_uid

        if includeChildren.lower() == 'false':
            url = url + '?compact=true'

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def post_tasks(self, url, session, post_dict):
        '''
        Handle action post
        '''


        url = url + '/tasks'

        payload = post_dict

        try:
            resp = resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def put_tasks(self, url, session, job_uuid, action):
        '''
        Handle action cancel
        '''

        url = url + '/tasks'

        job_list = job_uuid.split(',')

        payload = {'action':action, 'list':job_list}
        try:
            resp = resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def delete_tasks(self, url, session, job_uuid):
        '''
        Handle action delete
        '''

        url = url + '/tasks/'+job_uuid

        try:
            resp = resp = session.delete(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp


    def put_tasks_update(self, url, session, updated_dict):
        '''
        Handle action update
        '''


        url = url + '/tasks' + '/' + updated_dict[0]['jobUID']

        payload = updated_dict[0]

        try:
            resp = resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_set_manifests(self,url, session, sol_id, filepath):
        resp = None
        param_dict = dict()
        url = url + '/manifests'
        
        try:
            
            if sol_id:
                url = url + '/' + str(sol_id)
            else:
                raise Exception("Invalid argument ID")
            
            if  filepath:                
                param_dict['filepath'] = filepath
                
                payload = dict()
                payload = param_dict
                resp = session.post(url,data = json.dumps(payload),verify=False, timeout=REST_TIMEOUT)
            else:
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
                
            resp.raise_for_status()
        except HTTPError as re:
            raise re

#################

    def list_resourcegroups(self, url, session, uuid):
        url = url + '/resourceGroups'

        try:

            if uuid:
                # If it is modify group request
                url = url + '/' + uuid

            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def criteriaproperties_resourcegroups(self, url, session):
        url = url + '/resourceGroups/criteriaProperties'

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def delete_resourcegroups(self, url, session, uuid):
        url = url + '/resourceGroups'

        try:
            url = url + '/' + uuid
            resp = session.delete(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def dynamic_resourcegroups(self, url, session, uuid, name, desc, type, criteria):
        resp = None
        url = url + '/resourceGroups'

        try:

            param_dict = dict()
            param_dict['name'] = name
            param_dict['description'] = desc
            param_dict['type'] = type
            param_dict['criteria'] = criteria

            if uuid:
                # If it is modify group request
                param_dict['uuid'] = uuid
                resp = session.put(url, data=json.dumps(param_dict), verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp
            elif name:
                resp = session.post(url, data=json.dumps(param_dict), verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def solution_resourcegroups(self, url, session, uuid, name, desc, type, solutionVPD, members, criteria):
        resp = None
        url = url + '/resourceGroups'

        try:

            if uuid:
                # If it is modify group request
                url = url + '/' + uuid
                if members:
                    payload = []
                    for dev in members:
                        param_dict = dict()
                        param_dict['op'] = 'add'
                        param_dict["path"] = "/members/-"
                        param_dict["value"] = dev
                        payload.append( param_dict )

                    resp = session.patch(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                    resp.raise_for_status()
                    return resp
            elif name:
                param_dict = dict()
                param_dict['name'] = name
                param_dict['description'] = desc
                param_dict['type'] = type
                
                #Validate if correct Solution VPD is set
                if  type == 'solution' and \
                    isinstance(solutionVPD, dict) and \
                    set(["id","machineType","model","serialNumber","manufacturer"]).issubset(set(solutionVPD.keys())):
                    param_dict['solutionVPD'] = solutionVPD
                else:
                    raise ValueError("Invalid Argument SolutionVPD")    
                
                param_dict['members'] = members
                param_dict['criteria'] = None
            
                payload = dict()
                payload = param_dict
                
                resp = session.post(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp
    
#################

    def list_osimage(self, url, session):
        resp        = None
        url         = url + '/osImages'

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def osimage_globalsettings(self, url, session, kwargs):
        resp        = None
        url         = url + '/osdeployment/globalSettings'
        try:
            # put call for globalSettings DONE
            if kwargs:
                if set(['activeDirectory', 'credentials', 'ipAssignment', 'isVLANMode', 'licenseKeys']).difference(
                        set(kwargs.keys())):
                    raise Exception(
                        "Invalid Arguments, Try:['activeDirectory'=<list>, 'credentials'=<list>,'ipAssignment','isVLANMode','licenseKeys'=<dict>]")
                if not isinstance(kwargs['activeDirectory'], dict) or \
                        not isinstance(kwargs['credentials'], list) or \
                        not isinstance(kwargs['licenseKeys'], dict):
                    raise Exception(
                        "Invalid Arguments, Try:['activeDirectory'=<list>, 'credentials'=<list>,'licenseKeys'=<dict>]")
                payload = {}
                for k, v in list(kwargs.items()):
                    payload[k] = v
                resp = self.put_method(url, session, payload)
                return resp
            else:
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
                resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def create_osimage_hostsettings(self, url, session, hosts):
        resp        = None
        url         = url + '/osdeployment/hostSettings'
        try:

            # post call for hostSettings DONE
            if hosts:
                for kwargs in hosts:
                    if set(['uuid', 'storageSettings', 'networkSettings']).difference(
                            set(kwargs.keys())):
                        raise Exception(
                            "Invalid Arguments, Try:[{'storageSettings'=<dict>, 'networkSettings'=<dict>,'uuid'}]")
                    if not isinstance(kwargs['storageSettings'], dict) or \
                            not isinstance(kwargs['networkSettings'], dict):
                        raise Exception(
                            "Invalid Arguments, Try:['storageSettings'=<dict>, 'networkSettings'=<dict>]")
                payload = hosts
                resp = self.post_method(url, session, payload)
                return resp
            else:
                raise Exception(
                    "Invalid Arguments, Try:[{'storageSettings'=<dict>, 'networkSettings'=<dict>,'uuid'}]")
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def update_osimage_hostsettings(self, url, session, hosts):
        resp        = None
        url         = url + '/osdeployment/hostSettings'
        try:

            # put call for globalSettings DONE
            if hosts:
                for kwargs in hosts:
                    if set(['uuid', 'storageSettings', 'networkSettings']).difference(
                            set(kwargs.keys())):
                        raise Exception(
                            "Invalid Arguments, Try:[{'storageSettings'=<dict>, 'networkSettings'=<dict>,'uuid'}]")
                    if not isinstance(kwargs['storageSettings'], dict) or \
                            not isinstance(kwargs['networkSettings'], dict):
                        raise Exception(
                            "Invalid Arguments, Try:['storageSettings'=<dict>, 'networkSettings'=<dict>]")
                payload = hosts
                resp = self.put_method(url, session, payload)
                return resp
            else:
                raise Exception(
                    "Invalid Arguments, Try:[{'storageSettings'=<dict>, 'networkSettings'=<dict>,'uuid'}]")
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def list_osimage_hostsettings(self, url, session, kwargs):
        resp        = None
        url         = url + '/osdeployment/hostSettings'

        if kwargs and 'uuid' in kwargs:
            url = url + '/' + kwargs['uuid']
        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def delete_osimage_hostsettings(self, url, session, kwargs):
        resp        = None
        url         = url + '/osdeployment/hostSettings'

        if kwargs and 'uuid' in kwargs:
            url = url + '/' + kwargs['uuid']
        else:
            raise Exception(
                "Invalid Arguments, Try:uuid is required")
        try:
            resp = session.delete(url, verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def osimage_hostplatforms(self, url, session, kwargs):
        resp        = None
        url         = url + '/hostPlatforms'
        try:
            if kwargs:
                if set(['networkSettings', 'selectedImage', 'storageSettings', 'uuid', ]).difference(
                        set(kwargs.keys())):
                    raise Exception(
                        "Invalid Arguments, Try:['networkSettings'=<dict>, 'selectedImage', 'storageSettings'=<dict>,'uuid',]")
                if not isinstance(kwargs['networkSettings'], dict) or not isinstance(kwargs['storageSettings'], dict):
                    raise Exception("Invalid Arguments, Try: networkSettings=<dict>, and storageSettings=<dict>")
                payload = {}
                for k, v in list(kwargs.items()):
                    payload[k] = v
                resp = self.put_method(url, session, [payload])
                return resp
            else:
                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
                resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def osimage_import(self, url, session, kwargs):
        resp        = None
        try:
            # postcall of osimage DONE
            payload = {}
            if 'imageType' in kwargs and 'jobId' not in kwargs:
                url = url + '/osImages'
                if kwargs['imageType'] not in ['BUNDLE', 'BUNDLESIG', 'BOOT', 'DUD', 'OS', 'OSPROFILE', 'SCRIPT', 'CUSTOM_CONFIG', 'UNATTEND']:
                    raise Exception("Invalid Arguments, Try: [BUNDLE,BUNDLESIG,BOOT,DUD,OS,OSPROFILE,SCRIPT, 'CUSTOM_CONFIG', 'UNATTEND']")
                if 'fileSize' in kwargs:
                    payload['fileSize'] = kwargs['fileSize']
                url = url + '/?imageType=' + kwargs['imageType']
                payload['Action'] = 'Init'
                resp = self.post_method(url, session, payload)
                return resp
            # postcall for jobID DONE
            if 'jobId' in kwargs:
                url = url + '/files/osImages'
                url = url + '?'
                if kwargs['imageType'] in ['BOOT', 'DUD', 'OS', 'OSPROFILE', 'SCRIPT', 'CUSTOM_CONFIG', 'UNATTEND']:
                    if set(['jobId','imageName','imageType','os']).difference(set(kwargs.keys())):
                        raise Exception ("Invalid Arguments, Try:['jobId','imageName','imageType','os']")
                elif kwargs['imageType'] in ['BUNDLE', 'BUNDLESIG']:
                    if set(['jobId', 'imageName', 'imageType']).difference(set(kwargs.keys())):
                        raise Exception("Invalid Arguments, Try:['jobId','imageName','imageType']")
                    ext = os.path.splitext(os.path.basename(kwargs['imageName']))[-1]
                    if not (ext in ['.zip','.asc']):
                        raise Exception("Invalid Arguments, Try:imageName extension with ['.asc','.zip'")
                if kwargs['imageType'] in ['BOOT', 'DUD'] and 'osrelease' not in kwargs :
                    raise Exception("Invalid Arguments, Try:['jobId','imageName','imageType','os','osrelease]")
                if 'serverId' in kwargs:
                    payload_keylist = ['serverId', 'path']
                    for k,v in  list(kwargs.items()):
                        if k in payload_keylist:
                            payload[k] = v
                        else:
                            url = url + "%s=%s&" %(k,v)
                    url = url.rstrip('&')
                    resp = session.post(url, data=json.dumps(payload), verify=False, timeout=100 * REST_TIMEOUT) # 15 minutes
                    return resp
                else:    # local case
                    if not kwargs['file']:
                        raise Exception("Invalid Arguments, file is required for local import")
                    url = url + "jobId="+str(kwargs['jobId'])+"&"
                    kwargs.pop('jobId')
                    #urllib.urlencode(kwargs['description'])
                    for k in  ['imageName', 'imageType', 'os', 'description']:
                        if k in kwargs:
                            url = url + "%s=%s&" %(k,kwargs[k])

                    if 'osrelease' in kwargs:
                        url = url + "%s=%s&" % ('osrelease', kwargs['osrelease'])
                    url = url.rstrip('&')
                    url = urllib.parse.quote(url, safe="%/:=&?~#+!$,;'@()*[]")
                    m = MultipartEncoder(
                        fields={ 'name':'uploadedfile', 'uploadedfile': (os.path.basename(kwargs['file']), open(kwargs['file'], 'rb'), 'application/octet-stream')}
                    )

                    logger.debug("before sending second call")
                    resp = session.post(url, data = m, headers={'Content-Type': m.content_type}, verify=False, timeout=100 * REST_TIMEOUT)
                    logger.debug("after  second call response")
                    return resp

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def osimage_remotefileservers(self, url, session, kwargs):
        resp = None
        try:
            payload = {}
            url = url + '/osImages/remoteFileServers'
            # put/delete call for remoteFileServers DONE

            if 'putid' in kwargs or 'deleteid' in kwargs:
                if 'putid' in kwargs:  # put call for remoteFileServers/<id>
                    url = url + '/' + kwargs['putid']
                    kwargs.pop('putid')
                    if set(['address', 'displayName', 'port', 'protocol']).difference(set(kwargs.keys())):
                        raise Exception("Invalid Arguments, Try:['address','displayName','port', 'protocol']")
                    for k, v in list(kwargs.items()):
                        payload[k] = v

                    resp = self.put_method(url, session, payload)
                    return resp

                if 'deleteid' in kwargs:  # delete call for remoteFileServers/<id>
                    if list(kwargs.keys()).__len__() != 1:
                        raise Exception("Invalid Arguments, Try:deleteid=<id> only")
                    url = url + '/' + kwargs['deleteid']
                    payload = {}
                    resp = self.delete_method(url, session, payload)
                    return resp
            elif 'address' in kwargs and 'putid' not in kwargs and 'deleteid' not in kwargs:
                # postcall for remoteFileServers DONE
                if set(['address', 'displayName', 'port', 'protocol']).difference(set(kwargs.keys())):
                    raise Exception("Invalid Arguments, Try:['address','displayName','port', 'protocol']")
                for k, v in list(kwargs.items()):
                    payload[k] = v
                resp = self.post_method(url, session, payload)
                return resp
            else:
                if 'id' in kwargs:
                    url = url + '/' + kwargs['id']

                resp = session.get(url, verify=False, timeout=REST_TIMEOUT)  ## It raises HTTPError here
                resp.raise_for_status()
                return resp

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def osimage_delete(self, url, session, kwargs):
        resp = None
        url = url + '/osImages'
        try:
            if 'id' not in kwargs:
                raise Exception("Invalid Arguments, Try: id='id1,id2, .. ,idn' ")
            url = url + '/' + str(kwargs['id'])
            resp = self.delete_method(url, session, payload={})
            return resp

        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_method(self, url, session, **kwargs):
        resp = None
        try:
            resp = session.get(url,verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re  ## uncomment this
        return resp

    def put_method(self, url, session, payload, **kwargs):
        resp = None
        try:
            resp = session.put(url, data = json.dumps(payload), verify=False, timeout=60)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re  ## uncomment this
        return resp

    def delete_method(self,url, session, payload, **kwargs):
        resp = None
        try:
            resp = session.delete(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re  ## uncomment this
        return resp

    def post_method(self,url, session, payload, **kwargs):
        resp = None
        try:
            resp = session.post(url, data = json.dumps(payload), verify=False, timeout=REST_TIMEOUT)    ## It raises HTTPError here
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re  ## uncomment this
        return resp

    def get_rules(self,url, session, id):
        url = url + '/compliance/rules'

        if id:
            url = url + '/' + id

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def set_rules(self, url, session, rule):
        resp = None
        url = url + '/compliance/rules'

        try:
            resp = session.post(url, data=json.dumps(rule), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def get_compositeResults(self,url, session, id, query_solutionGroups):
        url = url + '/compliance/compositeResults'

        if id:
            url = url + '/' + id
        elif query_solutionGroups:
            url = url + '?groupID=' + query_solutionGroups

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def set_compositeResults(self, url, session, solutionGroups, targetResources, all_rules):
        resp = None
        url = url + '/compliance/compositeResults'

        try:
            if all_rules:
                resp = session.post(url, verify=False, timeout=REST_TIMEOUT)
                resp.raise_for_status()
                return resp

            payload = dict()
            if solutionGroups:
                payload['solutionGroups'] = solutionGroups
            elif targetResources:
                payload['targetResources'] = targetResources

            resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def get_storedcredentials(self,url, session, id):
        """
        Get stored credential details
        :param url:
        :param session:
        :param id:
        :return:
        """
        url = url + '/storedCredentials'

        if id:
            url = url + '/' + id

        try:
            resp = session.get(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def post_storedcredentials(self, url, session, user_name, password, description):
        """
        Add new credentials to store
        :param url:
        :param session:
        :param user_name:
        :param password:
        :param description:
        :return:
        """
        resp = None
        url = url + '/storedCredentials'

        try:
            payload = dict()
            payload['userName'] = user_name
            payload['password'] = password
            payload['description'] = description
            resp = session.post(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def put_storedcredentials(self, url, session, id, user_name, password, description):
        """
        Update already stored credential with specified id
        :param url:
        :param session:
        :param id:
        :param user_name:
        :param password:
        :param description:
        :return:
        """
        resp = None
        url = url + '/storedCredentials'

        if id:
            url = url + '/' + id
        else:
            raise Exception("Invalid Arguments, id is mandatory for put operation only")

        if not user_name or not password:
            raise Exception("Invalid Arguments, userName and password are mandatory for put operation only")

        try:
            payload = dict()
            payload['userName'] = user_name
            payload['password'] = password
            payload['description'] = description
            resp = session.put(url, data=json.dumps(payload), verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            raise re

        return resp

    def delete_storedcredentials(self, url, session, delete_id):
        """
        delete stored credential with specified id.
        :param url:
        :param session:
        :param delete_id:
        :return:
        """
        url = url + '/storedCredentials'

        if id:
            url = url + '/' + delete_id

        try:
            resp = session.delete(url, verify=False, timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp

    def get_license(self, url, session):
        url = url + '/notificationsLicense'

        try:
            resp = session.get(url,verify=False,timeout=REST_TIMEOUT)
            resp.raise_for_status()
        except HTTPError as re:
            logger.error("REST API Exception: Exception = %s", re)
            raise re
        return resp    
